
package net.tinyos.false_pos;

import net.tinyos.util.*;
import net.tinyos.message.*;
import java.io.*;
import java.net.*;
import java.util.*;
//import java.awt.event.*;

public class FalsePosComm implements MessageListener {	

	private static final String STATS_PRE = "FalsePosComm";

	private static final short NO_CTRL = 0;
	private static final short SET_VALS = 1;

	private static final int TOS_UART_ADDR = 0x7E;

	// TIMEOUT is in ms
	private static final int TIMEOUT = ((1000 * 60) * 30);

	Timer timer_;

	short senseTime_, runSeqno_;
	String fileName_;
	int seed_;

	MoteIF mote_;

	Vector nodeStats_;
    
	public static void main(String[] args) throws IOException {
		new FalsePosComm(args);
	}

	public FalsePosComm(String args[]) {

		senseTime_ = 0;
		seed_ = 2006;
		runSeqno_ = 1;
		fileName_ = new String();
		boolean fileNameWasSet = false;

		nodeStats_ = new Vector();

	try {
		// Parse the input options
		for(int i = 0; i < args.length; i++) {
			if(args[i].equals("-help") || args[i].equals("-h")) {
				help(null);
			} else if(args[i].equals("-senseTime")) {
				senseTime_ = Short.parseShort(args[i+1]);
				i++;	
			} else if(args[i].equals("-seqno")) {
				runSeqno_ = Short.parseShort(args[i+1]);
				i++;	
			} else if(args[i].equals("-seed")) {
				seed_ = Integer.parseInt(args[i+1]);
				i++;
			} else if(args[i].equals("-file")) {
				fileName_ = args[i+1];
				fileNameWasSet = true;
				i++;
			} else {
				help("unrecongized option: " + args[i]);
			}
		}

			if(!fileNameWasSet) {
				fileName_ = STATS_PRE + "-sense" + senseTime_ +
				"-seqno" + runSeqno_ + "-seed" + seed_ + ".stats";
			}

	} catch (Exception e) {
		e.printStackTrace();
		help("exception parsing args: " + e);
		return;
	}

		// Connect to the serial forwarder and listen for pkts
		mote_ = new MoteIF();
		mote_.registerListener(new StatsMsg(), this);
		mote_.registerListener(new UartCtrlMsg(), this);

		// Send the parameters to the mote and let TinyOS
		// do its thing until we receive a stats packet
		sendParameters();

		// Start a timer to kill this program just in case
		// we never get a terminate signal from the mote
		timer_ = new Timer();
		timer_.schedule(new ExitTask(), TIMEOUT);

	}

	private class ExitTask extends TimerTask {
		public void run() {
			System.exit(1);
		}
	}

	private void sendParameters() {

		UartCtrlMsg msg = new UartCtrlMsg();

		msg.set_ctrlType(SET_VALS);
		msg.set_senseTimeVal(senseTime_);
		msg.set_runSeqno(runSeqno_);

	try {
		System.err.println("sendMsg");
		System.err.println(msg.toString());
		mote_.send(MoteIF.TOS_BCAST_ADDR, msg);
	} catch (IOException ioe) {
		System.err.println(
			"Warning: Got IOException sending control message: "
			+ioe);
		ioe.printStackTrace();
	}

	}

	// We got a message from the mote
	public void messageReceived(int destAddr, Message msg) {
		System.err.println("msgRecv");
		if (msg instanceof StatsMsg) {
			statsReceived(destAddr, (StatsMsg)msg);
			int exitVal = writeStatsToFile();
			System.exit(exitVal);
		} else {
			throw new RuntimeException(
				"messageReceived: Got bad message type: "+msg);
		}
	}

	private void statsReceived(int destAddr, 
			StatsMsg msg) {

		NodeStats ns = new NodeStats();

		System.out.println(msg.toString());

		// Compare seqno with runSeqno
		if(msg.get_runSeqno() == runSeqno_) {

			// Collect the received stats in our array
			ns.setNodeId(msg.get_nodeId());
			ns.setFracIdle(msg.get_fracIdle());
			ns.setFracBusy(msg.get_fracBusy());
			ns.setRunSeqno(msg.get_runSeqno());

			nodeStats_.add(ns);
		}

	}

	private int writeStatsToFile() {

		int retVal = 1;

		// Sort our list based on node id
		Collections.sort(nodeStats_, new NodeIdComp());

	try {
		// Name the file based on the parameters
		PrintStream out = 
				new PrintStream(new FileOutputStream(fileName_));

		out.println("");
		out.println("Stats for senseTime=" + senseTime_ + 
				" (seed=" + seed_ + 
				", seqno=" + runSeqno_ + ")");
		out.println("");
		NodeStats ns;
		for(int i = 0; i < nodeStats_.size(); i++) {
			ns = (NodeStats)nodeStats_.elementAt(i);
			out.println("Node " + ns.getNodeId());
			out.println("\tFracIdle " + ns.getFracIdle());
			out.println("\tFracBusy " + ns.getFracBusy());
		}
		out.println("");
		out.close();

	} catch (IOException ioe) {
		System.err.println(
			"Warning: Got IOException opening output file: "+ioe);
		ioe.printStackTrace();

	}

		return retVal;
	}

	private void help(String error) {
		int exitStatus = 0;
		if(error != null) {
			System.err.println("Error parsing args (" + 
				error + ")\n");
			exitStatus = 1;
		}
		System.err.println("FalsePosComm usage:");
		System.err.println("\tjava net.tinyos.simple_pbbf.FalsePosComm" +
			" [options]");
		System.err.println("");
		System.err.println("Options:");
		System.err.println("\t-senseTime b-mac sensing time " +
			"(in ms, max = 255)");
		System.err.println("\t-seqno sequence number of the test");
		System.err.println("\t-seed random seed");
		System.err.println("\t-file output file name");
		System.err.println("");
		System.exit(exitStatus);
	}

} // end FalsePosComm class

class NodeIdComp implements Comparator {
	public int compare (Object o1, Object o2) {
		int n1 = ((NodeStats)o1).getNodeId();
		int n2 = ((NodeStats)o2).getNodeId();
		return (Math.abs(n1) - Math.abs(n2));
	}
} // end NodeIdComp class

class NodeStats {
	
	int nodeId_;
	short fracIdle_;
	short fracBusy_;
	short runSeqno_;

	public NodeStats() {
		nodeId_ = -1;
		fracIdle_ = -1;
		fracBusy_ = -1;
		runSeqno_ = -1;
	}

	public int getNodeId() {
		return nodeId_;
	}

	public void setNodeId(int v) {
		nodeId_ = v;
	}

	public short getFracIdle() {
		return fracIdle_;
	}

	public void setFracIdle(short v) {
		fracIdle_ = v;
	}

	public short getFracBusy() {
		return fracBusy_;
	}

	public void setFracBusy(short v) {
		fracBusy_ = v;
	}

	public short getRunSeqno() {
		return runSeqno_;
	}

	public void setRunSeqno(short v) {
		runSeqno_ = v;
	}

} // end NodeStats class


